from dataclasses import dataclass, field
from typing import Any, List, Dict

from .errors import ParseError

import jwt
import bson
import traceback
import inspect

from dateutil.parser import parse as parseDate


@dataclass
class Field():
    optional: bool = False
    missingStatusCode: int = None
    default: Any = None
    allowedTypes: List[str] = None
    description: str = ""
    parseFrom: str = None
    serializeTo: str = None

    def parser(self, value):
        return value

    def normalizer(self, value):
        return value

    def parse(self, value):
        value = self.parser(value)
        return self.normalize(value)

    def normalize(self, value):
        return self.normalizer(value)

    def serialize(self, value):
        return value

    def describe(self):
        return {
            "allowedTypes": self.allowedTypes,
            "optional": self.optional,
            "default": self.default,
            "description": self.description,
        }

    def __repr__(self):
        return "FIELD"


@dataclass
class StringField(Field):
    allowedTypes = ["String"]

    def parser(self, value):
        value = super().parser(value)
        if not isinstance(value, str):
            raise ParseError(message="\"{}\" is not a string".format(value))
        return value


@dataclass
class NoWhitespaceStringField(StringField):
    """
        A string field where leading/trailing whitespace
        should be trimmed.
    """

    def normalizer(self, value):
        value = super().normalizer(value)
        return value.strip()


@dataclass
class NonBlankStringField(NoWhitespaceStringField):
    def parser(self, value):
        value = super().parser(value)
        if not value:
            raise ParseError(message="Cannot be blank")
        return value


@dataclass
class EmailField(NonBlankStringField):
    def parser(self, value):
        value = super().parser(value)
        if "@" not in value:
            raise ParseError(
                message="\"{}\" is not a valid email address".format(value)
            )
        return value


@dataclass
class ObjectIdField(Field):
    allowedTypes = ["String"]

    def parser(self, value):
        try:
            return bson.ObjectId(value)
        except:
            raise ParseError(message="\"{}\" is not a valid id".format(value))
        return value

    def serialize(self, value):
        return str(value)


@dataclass
class DatetimeField(Field):
    allowedTypes = ["String"]

    def parser(self, value):
        try:
            return parseDate(value)
        except:
            raise ParseError(
                message="\"{}\" is not a valid datetime".format(value)
            )
        return value

    def serialize(self, value):
        if value:
            return value.isoformat()
        return None


@dataclass
class BoolField(Field):
    allowedTypes = ["Boolean"]

    def parser(self, value):
        try:
            return value == True
        except:
            raise ParseError(
                message="\"{}\" is not a valid boolean".format(value)
            )
        return value


@dataclass
class JwtField(Field):
    secretKey: str = ""

    def parser(self, value):
        value = super().parser(value)
        if not value:
            return None
        parsed = jwt.decode(
            value,
            self.secretKey,
            algorithms=['HS256'],
        )
        return parsed


@dataclass
class DictField(Field):
    def __init__(self, fields: dict, *args, **kwargs):
        super().__init__(*args, **kwargs)
        fields = fields or {}
        for k, v in fields.items():
            setattr(self, k, v)

    def _getFields(self):
        return {
            f: theField
            for (f, theField) in filter(
                lambda k: isinstance(k[1], Field),
                inspect.getmembers(self),
            )
        }

    def parser(self, source: Dict):
        source = source or {}
        fields = self._getFields()
        if not fields:
            return source
        parameters = {}
        for name, parameter in fields.items():
            if (not parameter.optional) and (name not in source):
                raise ParseError(
                    message="Missing required field: {}".format(name),
                    statusCode=parameter.missingStatusCode,
                )
            parameters[name] = parameter.parse(
                source.get(name, parameter.default)
            )
        return parameters

    def serialize(self, dictValue):
        fields = self._getFields()
        if not fields:
            return dictValue
        result = {}

        for name, outField in fields.items():
            try:
                value = dictValue[name]

                name = getattr(outField, "serializeTo", name) or name

                try:
                    result[name] = outField.serialize(value)

                except:
                    result[name] = value
            except KeyError as k:
                if not outField.optional:
                    print("MISSING FIELD FOR SERIALIZATION", k)
                    result[name] = None

        return result

    def describe(self):
        return {
            key: field.describe()
            for (key, field) in self._getFields().items()
        }


@dataclass
class ErrorField(DictField):
    message = StringField()


@dataclass
class ListField(Field):
    def __init__(self, elementField):
        self.elementField = elementField
        self.default = []

    def parser(self, source):
        if not source:
            return []
        parsed = []
        for value in source:
            parsed.append(elementField.parser(value))
        return parsed

    def serialize(self, values):
        result = []
        if not values:
            return result
        for value in values:
            try:
                result.append(self.elementField.serialize(value))
            except Exception as e:
                print(e)
                result.append(value)

        return result


from aiohttp import web
from typing import Dict
import json
from .fields import Field, ListField, ErrorField
from .errors import ParseError, ClientError
import datetime
import inspect


class DictParser():
    def __init__(self, settings=None, *args, **kwargs):
        self.settings = settings or {}
        return

    async def _parse_body(self, request):
        try:
            return json.loads(await request.text())
        except Exception as e:
            print(e)
            return {}

    def _getFields(self):
        return {
            f: theField
            for (f, theField) in filter(
                lambda k: isinstance(k[1], Field),
                inspect.getmembers(self),
            )
        }

    async def parse(self, source: Dict):
        parameters = {}
        for name, parameter in self._getFields().items():
            if isinstance(parameter, Field):
                if (not parameter.optional) and (name not in source):
                    raise ParseError(
                        message="Missing required field: {}".format(name),
                        statusCode=parameter.missingStatusCode,
                    )
                parameters[name] = parameter.parse(
                    source.get(name, parameter.default)
                )
        return parameters

    def describe(self):
        return {
            key: field.describe()
            for (key, field) in self._getFields().items()
        }


class RequestParser():
    BodyParser = DictParser
    QueryParser = DictParser
    CookiesParser = DictParser
    HeadersParser = DictParser
    UrlParser = DictParser

    def __init__(self, *args, **kwargs):
        self.settings = kwargs.get("settings", None)
        return

    async def _parse_body(self, request):
        try:
            if request.can_read_body:
                return await request.json()
            return None
        except Exception as e:
            print(e)
            raise ParseError(message="Unable to parse body")

    async def parse(self, request):
        parameters = {}

        if self.BodyParser:
            body = await self._parse_body(request)
            if body:
                parameters.update(await self.BodyParser().parse(body))

        if self.QueryParser:
            parameters.update(await self.QueryParser().parse(request.query))

        if self.HeadersParser:
            parameters.update(
                await self.HeadersParser().parse(request.headers)
            )

        if self.CookiesParser:
            parameters.update(
                await self.CookiesParser(settings=self.settings, ).parse(
                    request.cookies
                )
            )

        try:
            parameters.update(
                await self.UrlParser(settings=self.settings, ).parse(
                    request.match_info
                )
            )
        except:
            raise ClientError(statusCode=404)

        return parameters

    def describe(self):
        return {
            "body": self.BodyParser().describe(),
        }


class ResponseSerializer():
    def __init__(self, *args, **kwargs):
        return

    def _getFields(self):
        return {
            f: self.__getattribute__(f)
            for f in dir(self)
            if isinstance(self.__getattribute__(f), Field)
        }

    async def serialize(
        self,
        result,
        statusCode=200,
        headers=None,
        cookies=None,
        deletedCookies=None,
    ):
        body = {}
        fields = self._getFields()
        for name, resultField in fields.items():

            if hasattr(resultField, "default"):
                value = result.get(name, resultField.default)
            else:
                value = result[name]

            try:
                value = resultField.serialize(value)
            except Exception as e:
                print(e)
                pass

            serializeTo = getattr(resultField, "serializeTo", name) or name
            body[serializeTo] = value

        headers = {k: "; ".join(vs) for k, vs in headers.items()}

        response = web.json_response(
            body,
            status=statusCode,
            headers=headers or {},
        )

        for c in (cookies or {}).values():

            try:
                value = c["value"].decode("utf")
            except:
                value = c["value"]
            response.set_cookie(
                c["name"],
                value,
                secure=c["secure"],
                httponly=c["httpOnly"],
                expires=c.get(
                    "expires",
                    datetime.datetime.utcnow() + datetime.timedelta(days=1)
                )
            )

        for c in (deletedCookies or {}).values():
            response.del_cookie(c["name"], )

        return response

    def describe(self):
        return {
            "body": {
                key: field.describe()
                for (key, field) in self._getFields().items()
            }
        }


class ResponseSerializerWithErrors(ResponseSerializer):
    errors = ListField(ErrorField)


import sys
sys.path.append(".")

from .. import Clericus
from ..config import defaultSettings, connectToDB

import unittest
from aiohttp.test_utils import AioHTTPTestCase, unittest_run_loop

import json

from ..schemas import createCollections


class ClericusTestCase(AioHTTPTestCase):
    async def tearDownAsync(self):
        await self.db.client.drop_database(self.db.name)

    async def get_application(self):
        settings = defaultSettings()
        settings["db"]["name"] = f"test{type(self).__name__}"
        settings = connectToDB(settings)
        await settings["db"].client.drop_database(settings["db"].name)
        self._settings = settings
        self.db = settings["db"]
        await createCollections(self.db)
        return Clericus(self._settings)


import motor
import bson
import asyncio
import bcrypt
import datetime

from pymongo.errors import DuplicateKeyError
from ..errors import ClientError
from ..errors.mongo import parseDuplicateKeyError
from ..db.transactions import withSession


async def createAuthenticationCollection(db):
    try:
        await db.create_collection("authentication")
    except:
        pass
    await db.command(
        "collMod",
        "authentication",
        validator={
            "$jsonSchema": {
                "bsonType": "object",
                "required": ["userID", "passwordHash", "lastUpdated"],
                "properties": {
                    "userID": {
                        "bsonType": "objectId",
                    },
                    "passwordHash": {
                        "bsonType": "binData",
                    },
                    "lastUpdated": {
                        "bsonType": "date",
                    }
                }
            }
        }
    )

    await asyncio.wait([
        db.authentication.create_index(
            [("userID", 1)],
            unique=True,
            name="uniqueUserID",
        ),
    ])


@withSession
async def setPassword(
    db,
    session,
    userID: bson.ObjectId,
    password: str,
):
    try:
        #bcrypt wants bytes
        password = password.encode("utf-8")
    except:
        pass
    authentication = {
        "userID": userID,
        "passwordHash": bcrypt.hashpw(password, bcrypt.gensalt()),
        "lastUpdated": datetime.datetime.utcnow().replace(
            tzinfo=datetime.timezone.utc
        )
    }
    result = await db.authentication.update_one(
        {"userID": userID},
        {
            "$set": authentication,
        },
        upsert=True,
        session=session,
    )
    # print(result.upserted_id)
    if result.upserted_id:
        authentication["_id"] = result.upserted_id
    elif result.matched_count == 1:
        authentication["_id"] = (
            await db.authentication.find_one({"userID": userID})
        )["_id"]
    else:
        raise Exception("Unable to set password")
    return authentication


@withSession
async def authenticateUser(
    db,
    session,
    email: str,
    password: str,
):
    try:
        #bcrypt wants bytes
        password = password.encode("utf-8")
    except:
        pass
    try:
        user = (
            await db.user.aggregate(
                pipeline=[
                    {
                        "$match": {
                            "email": email
                        }
                    },
                    # should be unnecessary
                    {
                        "$limit": 1
                    },
                    {
                        "$lookup": {
                            "localField": "_id",
                            "foreignField": "userID",
                            "from": "authentication",
                            "as": "authentication"
                        }
                    },
                    {
                        "$unwind": "$authentication"
                    },
                ],
                session=session,
            ).to_list(1)
        )[0]
    except:
        raise ClientError(
            message="Invalid Login",
            statusCode=401,
        )

    valid = bcrypt.checkpw(
        password,
        user["authentication"]["passwordHash"],
    )

    if not valid:
        raise ClientError(
            message="Invalid Login",
            statusCode=401,
        )

    # don't let the authentication info leave this function
    del user["authentication"]

    return user



import enum

from ..parsing.fields import ListField, ObjectIdField, DictField
from ..schemas.user import EmbeddedUserField


class Permissions(enum.Enum):
    NULL = None
    OWNER = "ownerIDs"
    VIEW = "viewerIDs"


def _toLookupKey(permission):
    if not permission.value.endswith("IDs"):
        raise Exception("Invalid permission type")
    return f"{permission.value[:-3]}s"


def permissionSelector(currentUser, permission):
    currentUserID = currentUser["_id"] if currentUser else None

    if permission == Permissions.VIEW:
        ors = [
            {
                "permissions.public": True
            },
        ]
        if currentUserID:
            ors += [
                {
                    f"permissions.{Permissions.OWNER.value}": currentUserID
                },
                {
                    f"permissions.{Permissions.VIEW.value}": currentUserID
                },
            ]
        return {"$or": ors}
    return {f"permissions.{permission.value}": currentUserID}


def lookupPermissionsStage(permissions: list):
    """
        Return a list of mongodb pipeline stages to lookup
        the given permission list on objects in the pipeline
    """
    pipeline = []
    for permission in permissions:
        pipeline += [
            {
                "$lookup": {
                    "from": "user",
                    "as": f"permissionUsers.{_toLookupKey(permission)}",
                    "localField": f"permissions.{permission.value}",
                    "foreignField": "_id",
                },
            },
        ]

    return pipeline


class PermissionField(DictField):
    def __init__(self, permissions, *args, **kwargs):
        super().__init__(
            fields={
                _toLookupKey(permission): ListField(EmbeddedUserField())
                for permission in permissions
            },
            *args,
            **kwargs,
        )


class Actions():
    def __init__(self, actions):
        self._actions = actions or {}

    def actions(self):
        return self._actions.keys()

    def canDoAction(
        self,
        action="",
        currentUser=None,
        targetObject=None,
        **kwargs,
    ):
        try:
            return self._actions[action](
                currentUser=currentUser,
                targetObject=targetObject,
                **kwargs,
            )
        except KeyError:
            raise Exception(f"Unknown Action: {action}")

    def checkActions(
        self,
        currentUser=None,
        targetObject=None,
        **kwargs,
    ):
        return {
            a: self.canDoAction(
                a,
                currentUser=currentUser,
                targetObject=targetObject,
                **kwargs,
            )
            for a in self.actions()
        }


def hasPermission(permission=Permissions.NULL, ):
    def check(
        currentUser=None,
        targetObject=None,
        **kwargs,
    ):
        try:
            return currentUser["_id"] in targetObject["permissions"][
                permission.value]
        except:
            return False

    return check


def hasAllPermissions(*permissions):
    def check(
        currentUser=None,
        targetObject=None,
        **kwargs,
    ):
        return all([
            hasPermission(p)(
                currentUser=currentUser,
                targetObject=targetObject,
                **kwargs,
            ) for p in permissions
        ])

    return check


def hasAnyPermissions(*permissions):
    def check(
        currentUser=None,
        targetObject=None,
        **kwargs,
    ):
        return any([
            hasPermission(p)(
                currentUser=currentUser,
                targetObject=targetObject,
                **kwargs,
            ) for p in permissions
        ])

    return check


def orClause(*clauses):
    def check(
        currentUser=None,
        targetObject=None,
        **kwargs,
    ):
        return any([
            clause(
                currentUser=currentUser,
                targetObject=targetObject,
                **kwargs,
            ) for clause in clauses
        ])

    return check


def isPublic(
    currentUser=None,
    targetObject=None,
    **kwargs,
):
    try:
        return targetObject["permissions"]["public"] == True
    except:
        return False


from pymongo.errors import DuplicateKeyError
from dataclasses import dataclass


@dataclass
class ParsedDuplicateKeyError:
    collectionName: str
    indexName: str


def parseDuplicateKeyError(error: DuplicateKeyError):
    """
        Parse an error message string from mongo where the
        error is formatted like:
            'E11000 duplicate key error collection: test.user index: uniqueEmail dup key: { : ";E7C/E1\nz1W)AG?1\x08-EA\x01\x17" }'
    """

    msg = error.details["errmsg"]

    prefix = "E11000 duplicate key error "

    if not msg.startswith(prefix):
        #this isn't parsable
        raise error

    try:
        # should be the second and fourth words after the prefix
        parts = msg[len(prefix):].split(" ")
        return ParsedDuplicateKeyError(
            collectionName=parts[1],
            indexName=parts[3],
        )
    except:
        raise error


import sys
sys.path.append(".")

import unittest
from aiohttp.test_utils import unittest_run_loop
from .test_case import ClericusTestCase

import faker
fake = faker.Faker()


class LoginTestCase(ClericusTestCase):
    @unittest_run_loop
    async def testLogin(self):
        resp = await self.client.request("GET", "/me/")
        # not logged in
        self.assertEqual(resp.status, 401)
        data = await resp.json()

        user = {
            "username": fake.name(),
            "email": fake.email(),
            "password": fake.password(),
        }
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        # sign up
        self.assertEqual(resp.status, 200)

        resp = await self.client.request("GET", "/me/")
        # logged in from signup
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data["currentUser"]["username"], user["username"])

        resp = await self.client.request("GET", "/log-out/")
        # log out
        self.assertEqual(resp.status, 200)
        data = await resp.json()

        # logged out from log-out
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 401)

        resp = await self.client.request(
            "POST",
            "/log-in/",
            json={
                "email": user["email"],
                "password": user["password"],
            },
        )
        # log in
        self.assertEqual(resp.status, 200)
        data = await resp.json()

        # logged in successfully
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 200)

    @unittest_run_loop
    async def testInvalidPassword(self):
        user = {
            "username": fake.name(),
            "email": fake.email(),
            "password": fake.password(),
        }
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        # sign up
        self.assertEqual(resp.status, 200)

        resp = await self.client.request("GET", "/log-out/")
        # log out
        self.assertEqual(resp.status, 200)

        resp = await self.client.request(
            "POST",
            "/log-in/",
            json={
                "email": user["email"],
                "password": user["password"] + "moo",
            },
        )
        # log in
        self.assertEqual(resp.status, 401)
        data = await resp.json()

        # logged in successfully
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 401)

    @unittest_run_loop
    async def testInvalidEmail(self):
        user = {
            "username": fake.name(),
            "email": fake.email(),
            "password": fake.password(),
        }
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        # sign up
        self.assertEqual(resp.status, 200)

        resp = await self.client.request("GET", "/log-out/")
        # log out
        self.assertEqual(resp.status, 200)

        resp = await self.client.request(
            "POST",
            "/log-in/",
            json={
                "email": fake.email(),
                "password": user["password"],
            },
        )
        # log in
        self.assertEqual(resp.status, 401)
        data = await resp.json()

        # logged in successfully
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 401)


if __name__ == '__main__':
    unittest.main()


from ..handler import Method, Endpoint
from ..parsing import (
    RequestParser,
    DictParser,
    ResponseSerializerWithErrors,
)
from ..parsing.fields import (
    StringField,
    Field,
    ListField,
    DictField,
    BoolField,
    ObjectIdField,
)


from ..schemas.permissions import (permissionSelector, lookupPermissionsStage, Permissions, PermissionField,)

from ..errors import ServerError, ClientError
import datetime
import bson


def generatePublishEndpoint(
    collectionName,
    permissionsList,
):
    class PublishEndpoint(Endpoint):
        class Post(Method):
            # class Parser(RequestParser):
            # class BodyParser(DictParser):
            #     objectID = ObjectIdField()

            class Serializer(ResponseSerializerWithErrors):
                permissions = DictField(
                    fields={
                        "_id": ObjectIdField(serializeTo="id"),
                        "permissionUsers": PermissionField(permissionsList),
                        "isPublic": BoolField(),
                    }
                )

            async def process(
                self,
                objectID: bson.ObjectId,
                currentUser,
            ):

                objectID = bson.ObjectId(objectID)

                selector = {
                    "$and": [
                        {
                            "_id": objectID
                        },
                        permissionSelector(
                            currentUser=currentUser,
                            permission=Permissions.OWNER,
                        ),
                    ]
                }

                pipeline = [
                    {
                        "$match": selector
                    },
                    {
                        "$addFields": {
                            "isPublic": {
                                "$eq": ["$permissions.public", True]
                            },
                        }
                    },
                ]
                pipeline += lookupPermissionsStage(permissionsList)

                items = await self.settings.db.get_collection(
                    collectionName
                ).aggregate(pipeline).to_list(1)

                if len(items) == 0:
                    raise ClientError(statusCode=404)

                result = await self.settings.db.get_collection(
                    collectionName
                ).update_one(
                    {"_id": objectID},
                    {"$set": {
                        "permissions.public": True
                    }},
                )

                items = await self.settings.db.get_collection(
                    collectionName
                ).aggregate(pipeline).to_list(1)

                return {"permissions": items[0]}

    return PublishEndpoint


def generateUnpublishEndpoint(
    collectionName,
    permissionsList,
):
    class UnpublishEndpoint(Endpoint):
        class Post(Method):
            # class Parser(RequestParser):
            # class BodyParser(DictParser):
            #     objectID = ObjectIdField()

            class Serializer(ResponseSerializerWithErrors):
                permissions = DictField(
                    fields={
                        "_id": ObjectIdField(serializeTo="id"),
                        "permissionUsers": PermissionField(permissionsList),
                        "isPublic": BoolField(),
                    }
                )

            async def process(
                self,
                objectID: bson.ObjectId,
                currentUser,
            ):

                objectID = bson.ObjectId(objectID)

                selector = {
                    "$and": [
                        {
                            "_id": objectID
                        },
                        permissionSelector(
                            currentUser=currentUser,
                            permission=Permissions.OWNER,
                        ),
                    ]
                }

                pipeline = [
                    {
                        "$match": selector
                    },
                    {
                        "$addFields": {
                            "isPublic": {
                                "$eq": ["$permissions.public", True]
                            },
                        }
                    },
                ]
                pipeline += lookupPermissionsStage(permissionsList)

                items = await self.settings.db.get_collection(
                    collectionName
                ).aggregate(pipeline).to_list(1)

                if len(items) == 0:
                    raise ClientError(statusCode=404)

                result = await self.settings.db.get_collection(
                    collectionName
                ).update_one(
                    {"_id": objectID},
                    {"$set": {
                        "permissions.public": False
                    }},
                )

                items = await self.settings.db.get_collection(
                    collectionName
                ).aggregate(pipeline).to_list(1)

                return {"permissions": items[0]}

    return UnpublishEndpoint


def generateActionsEndpoint(
    collectionName,
    actions,
):
    class CheckPermissionsEndpoint(Endpoint):
        class Get(Method):
            class Parser(RequestParser):
                class UrlParser(DictParser):
                    objectID = ObjectIdField()

            class Serializer(ResponseSerializerWithErrors):
                actions = DictField(fields={})

            async def process(
                self,
                objectID: bson.ObjectId,
                currentUser=None,
            ):

                targetObject = await self.settings.db.get_collection(
                    collectionName
                ).find_one(objectID)

                response = actions.checkActions(
                    currentUser=currentUser,
                    targetObject=targetObject,
                )

                return {"actions": response}

    return CheckPermissionsEndpoint


from ..handler import Method, Endpoint
from ..parsing import (
    RequestParser,
    DictParser,
    ResponseSerializerWithErrors,
)
from ..parsing.fields import (
    NonBlankStringField,
    EmailField,
    StringField,
    JwtField,
    Field,
    DictField,
)

from ..schemas.user import createUser

from ..schemas.authentication import authenticateUser
from ..middleware import setCurrentUser

from ..errors import ServerError
import jwt
import datetime
import bson


class SignUpEndpoint(Endpoint):
    """
        Handle Initial account creation
    """

    class Post(Method):
        """
            Create a new user with the given username, email, and password.
        """

        class Parser(RequestParser):
            class BodyParser(DictParser):
                email = EmailField(
                    description="The email of the user being created"
                )
                username = NonBlankStringField(
                    description="The username of the user being created"
                )
                password = NonBlankStringField(
                    description="The password to set for the new user"
                )

        class Serializer(ResponseSerializerWithErrors):
            def __init__(self):
                self.success = Field(
                    description="A boolean of whether the user was created"
                )

        async def process(
            self,
            email: str,
            username: str,
            password: str,
        ):

            user = await createUser(
                db=self.settings.db,
                username=username,
                email=email,
                password=password,
            )

            token = await setCurrentUser(
                self,
                self.settings.db,
                self.settings.jwtKey,
                user,
            )
            return {"success": "yes"}


class LogInEndpoint(Endpoint):
    class Post(Method):
        class Parser(RequestParser):
            class BodyParser(DictParser):
                email = EmailField()
                password = StringField()

        class Serializer(ResponseSerializerWithErrors):
            token = StringField()
            currentUser = DictField({
                "id": StringField(),
                "username": StringField(),
                "email": EmailField(),
            })

        async def process(
            self,
            email: str,
            password: str,
            # currentUser=None,
        ):
            user = await authenticateUser(
                db=self.settings.db,
                email=email,
                password=password,
            )
            token = await setCurrentUser(
                self,
                self.settings.db,
                self.settings.jwtKey,
                user,
            )

            return {
                "token": token,
                "currentUser": {
                    "id": str(user["_id"]),
                    "username": user["username"],
                    "email": user["email"],
                },
            }


class LogOutEndpoint(Endpoint):
    class Get(Method):

        # class Serializer(ResponseSerializerWithErrors):
        #     def __init__(self):
        #         self.token = StringField()

        async def process(
            self,
            authenticated: str = None,
        ):

            self.unsetCookie("authentication", )
            return {}


class MeEndpoint(Endpoint):
    class Get(Method):
        class Serializer(ResponseSerializerWithErrors):
            currentUser = DictField({
                "id": StringField(),
                "username": StringField(),
                "email": EmailField(),
            })

        async def process(self, currentUser):
            await setCurrentUser(
                self,
                self.settings.db,
                self.settings.jwtKey,
                currentUser,
            )
            return {
                "currentUser": {
                    "id": str(currentUser["_id"]),
                    "username": currentUser["username"],
                    "email": currentUser["email"],
                },
            }



import motor
import asyncio

from pymongo.errors import DuplicateKeyError
from ..errors import ClientError
from ..errors.mongo import parseDuplicateKeyError

from ..db.transactions import withSession

from .authentication import setPassword

from ..parsing.fields import (
    ObjectIdField,
    StringField,
    DictField,
    DatetimeField,
)


async def createUserCollection(db):
    try:
        await db.create_collection("user")
    except:
        pass
    await db.command(
        "collMod",
        "user",
        validator={
            "$jsonSchema": {
                "bsonType": "object",
                "required": ["username", "email"],
                "properties": {
                    "username": {
                        "bsonType": "string",
                        "minLength": 1,
                        "description": "username cannot be blank"
                    },
                    "email": {
                        "bsonType": "string",
                        "minLength": 1,
                        "description": "username cannot be blank"
                    },
                }
            }
        }
    )

    await asyncio.wait([
        db.user.create_index(
            [("email", 1)],
            unique=True,
            name="uniqueEmail",
            collation={
                "locale": "en",
                "strength": 2
            },
        ),
        db.user.create_index(
            [("username", 1)],
            unique=True,
            name="uniqueUsername",
            collation={
                "locale": "en",
                "strength": 2
            },
        ),
    ])


@withSession
async def createUser(
    db,
    username: str,
    email: str,
    password: str = None,
    session=None,
):
    user = {
        "username": username,
        "email": email,
    }
    try:
        result = await db.user.insert_one(
            user,
            session=session,
        )
        user["_id"] = result.inserted_id

        await setPassword(
            db=db,
            session=session,
            userID=user["_id"],
            password=password,
        )

        return user
    except DuplicateKeyError as err:
        parsedError = parseDuplicateKeyError(err)
        if parsedError.indexName == "uniqueEmail":
            raise ClientError(
                message=f"The email \"{email}\" is already in use",
                statusCode=422,
            )
        if parsedError.indexName == "uniqueUsername":
            raise ClientError(
                message=f"The username \"{username}\" is already in use",
                statusCode=422,
            )
        # unknown error
        raise err

    except Exception as err:
        raise err


class EmbeddedUserField(DictField):
    def __init__(self, *args, **kwargs):
        super().__init__(
            fields={
                "_id": ObjectIdField(serializeTo="id"),
                "username": StringField(),
                # "email": StringField(),
            },
            *args,
            **kwargs,
        )



from aiohttp import web
from aiohttp.web_middlewares import normalize_path_middleware
from dataclasses import dataclass, field
from types import SimpleNamespace as SN
from typing import Sequence
from .routes import (
    authentication as authenticationRoutes,
    permissions as permissionRoutes,
    health as healthRoutes
)
from .config import defaultSettings
from .middleware import logRequest, allowCors, authentication as authenticationMiddleware


class Clericus(web.Application):
    def __init__(self, settings=None):
        baseSettings = defaultSettings()
        baseSettings.update(settings or {})
        super().__init__(
            middlewares=[
                normalize_path_middleware(append_slash=True),
                allowCors(origins=baseSettings["corsOrigins"]),
                logRequest,
                authenticationMiddleware(
                    db=baseSettings["db"],
                    secretKey=baseSettings["jwtKey"],
                ),
            ]
        )

        self["settings"] = baseSettings

        self.add_endpoint(
            "/sign-up/",
            authenticationRoutes.SignUpEndpoint,
        )
        self.add_endpoint(
            "/log-in/",
            authenticationRoutes.LogInEndpoint,
        )
        self.add_endpoint(
            "/log-out/",
            authenticationRoutes.LogOutEndpoint,
        )
        self.add_endpoint(
            "/me/",
            authenticationRoutes.MeEndpoint,
        )
        self.add_endpoint(
            "/healthy/",
            healthRoutes.HealthCheckEndpoint,
        )

    def add_endpoint(self, path, handlerClass, name=None):
        self.router.add_route(
            "*", path,
            handlerClass(settings=SN(**self["settings"])).handle
        )
        self.router.add_route(
            "get", "/documentation{}".format(path),
            handlerClass(settings=SN(**self["settings"])).handleDocumentation
        )

    def run_app(self):
        web.run_app(self)




from ..parsing import RequestParser, ResponseSerializer
from aiohttp import web
from typing import ClassVar
import traceback
from inspect import getdoc
from .documentation import handleDocumentation
from ..errors import HTTPError, ClientError
import traceback
import inspect


class Method():
    def __init__(
        self,
        settings=None,
        statusCode=200,
        headers=None,
        cookies=None,
        deletedCookies=None,
    ):
        self.statusCode = statusCode
        self.headers = headers or {}
        self.cookies = cookies or {}
        self.deletedCookies = deletedCookies or {}
        self.settings = settings or {}

    class Parser(RequestParser):
        pass

    class Serializer(ResponseSerializer):
        pass

    async def parse(self, request):
        return await self.Parser(settings=self.settings).parse(request)

    async def process(self, **kwargs):
        return {}

    async def serialize(self, result):
        return await self.Serializer().serialize(
            result,
            statusCode=self.statusCode,
            headers=self.headers,
            cookies=self.cookies,
            deletedCookies=self.deletedCookies,
        )

    def setCookie(
        self,
        name,
        value,
        secure=False,
        httpOnly=False,
        expires=None,
    ):
        self.cookies[name] = {
            "name": name,
            "value": value,
            "secure": secure,
            "httpOnly": httpOnly,
            "expires": expires,
        }

    def unsetCookie(
        self,
        name,
    ):
        self.deletedCookies[name] = {
            "name": name,
        }

    def setHeader(self, name, value):
        if name in self.headers:
            self.headers[name].append(value)
        else:
            self.headers[name] = [value]

    async def handle(self, request: web.Request) -> web.Response:
        try:
            parsedData = await self.parse(request)
            parameters = inspect.signature(self.process).parameters
            if "currentUser" in parameters:
                # if the process function wants the current user, fill it
                # in from the authentication middleware
                currentUser = getattr(
                    request,
                    "currentUser",
                    None,
                )
                if all([
                    currentUser == None,
                    parameters["currentUser"].default ==
                    inspect.Parameter.empty,
                ]):
                    raise ClientError(
                        statusCode=401,
                        message="Unauthorized",
                        errorType="AuthenticationError",
                    )

                parsedData["currentUser"] = currentUser
            result = await self.process(**parsedData)
            return await self.serialize(result, )
        except HTTPError as e:
            return web.json_response(
                {"errors": [e.toJSON()]},
                status=e.statusCode,
            )
        except Exception as e:
            print(e)
            print(traceback.format_exc())
            return web.json_response(
                {
                    "errors": [{
                        "message": "Server Error",
                        "errorType": "Unknown",
                    }]
                },
                status=500,
            )

    def describe(self):
        return {
            "description": getdoc(self),
            "request": self.Parser().describe(),
            "response": self.Serializer().describe(),
        }


class Endpoint():
    def __init__(self, settings=None):
        self.settings = settings or {}
        self.Options = self.generateOptionsClass(
            list(self.methods().keys()) + ["options"],
            settings=self.settings,
        )

    def methods(self):
        return {
            f.lower(): self.__getattribute__(f)
            for f in dir(self)
            if isinstance(self.__getattribute__(f), Method.__class__)
            and f[0] != "_"
        }

    async def handle(self, request: web.Request) -> web.Response:
        method = self.methods().get(request.method.lower(), None)
        if method:
            return await method(settings=self.settings).handle(request)
        else:
            return await self.methodNotAllowedHandler(method)

    async def methodNotAllowedHandler(self, method):
        return web.json_response(
            {"errors": ["Method not allowed"]},
            status=405,
        )

    async def handleDocumentation(self, request: web.Request) -> web.Response:
        return await handleDocumentation(self)

    def generateOptionsClass(self, methods, settings=None):
        methodString = ", ".join([m.upper() for m in methods])

        settings = settings or {}

        class Options(Method):
            async def process(self, **kwargs):
                self.setHeader("Allow", methodString)

                # CORS
                self.setHeader("Access-Control-Request-Method", methodString)
                return

        return Options



import unittest
import asyncio

import sys
sys.path.append(".")
sys.path.append("..")
from ..parsing import DictParser
from ..parsing.fields import StringField, DictField


def async_test(f):
    def wrapper(self):
        return asyncio.run(f(self))

    return wrapper


class TestParsing(unittest.TestCase):
    @async_test
    async def test_moo(self):
        # self.assertEqual(1, 1)
        class dp(DictParser):
            stringOne = StringField()
            stringTwo = StringField()
            dictOne = DictField(
                fields={
                    "stringThree": StringField(),
                    "stringFour": StringField(),
                    "dictTwo": DictField({}),
                }
            )

        parsed = await dp().parse({
            "stringOne": "moo",
            "stringTwo": "cow",
            "dictOne": {
                "stringThree": "moo",
                "stringFour": "cow",
                "dictTwo": {
                    "x": 1
                }
            }
        })

        self.assertEqual(parsed["stringOne"], "moo")
        self.assertEqual(parsed["stringTwo"], "cow")
        self.assertEqual(parsed["dictOne"]["stringThree"], "moo")
        self.assertEqual(parsed["dictOne"]["stringFour"], "cow")
        self.assertEqual(parsed["dictOne"]["dictTwo"]["x"], 1)


if __name__ == '__main__':
    unittest.main()


import sys
sys.path.append(".")

import unittest
from aiohttp.test_utils import unittest_run_loop
from .test_case import ClericusTestCase

import faker
fake = faker.Faker()


class SignUpTestCase(ClericusTestCase):
    @unittest_run_loop
    async def testSignup(self):
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 401)
        data = await resp.json()
        user = {
            "username": fake.name(),
            "email": fake.email(),
            "password": fake.password(),
        }

        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 200)

        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data["currentUser"]["username"], user["username"])

    @unittest_run_loop
    async def testInvalidUsername(self):
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 401)
        data = await resp.json()
        user = {
            "username": "",
            "email": fake.email(),
            "password": fake.password(),
        }

        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

        user["username"] = 4
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

    @unittest_run_loop
    async def testInvalidEmail(self):
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 401)
        data = await resp.json()
        user = {
            "username": fake.name(),
            "email": "moo",
            "password": fake.password(),
        }

        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

        user["email"] = 5
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

    @unittest_run_loop
    async def testDuplicateUsernameSignup(self):
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 401)
        data = await resp.json()
        user = {
            "username": fake.name().lower(),
            "email": fake.email(),
            "password": fake.password(),
        }

        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 200)

        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data["currentUser"]["username"], user["username"])

        user["email"] = fake.email()
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

        # Test case insensitive matching
        user["username"] = user["username"].upper()
        user["email"] = fake.email()
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

        # Test with space
        user["username"] = "  " + user["username"] + "  "
        user["email"] = fake.email()
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

    @unittest_run_loop
    async def testDuplicateEmailSignup(self):
        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 401)
        data = await resp.json()
        user = {
            "username": fake.name(),
            "email": fake.email().lower(),
            "password": fake.password(),
        }

        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 200)

        resp = await self.client.request("GET", "/me/")
        self.assertEqual(resp.status, 200)
        data = await resp.json()
        self.assertEqual(data["currentUser"]["username"], user["username"])

        user["username"] = fake.name()
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

        # Test case insensitive matching
        user["email"] = user["email"].upper()
        user["username"] = fake.name()
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)

        # Test with space
        user["email"] = "    " + user["email"] + "   "
        user["username"] = fake.name()
        resp = await self.client.request(
            "POST",
            "/sign-up/",
            json=user,
        )
        self.assertEqual(resp.status, 422)
        data = await resp.json()
        self.assertEqual(len(data["errors"]), 1)


if __name__ == '__main__':
    unittest.main()


from aiohttp.web import middleware
import datetime
import bson
import jwt


@middleware
async def logRequest(request, handler):
    start = datetime.datetime.utcnow()
    resp = await handler(request)
    print(
        resp.status,
        request.method,
        request.path,
        getattr(
            request,
            "currentUser",
            {"username": "Unauthenticated"},
        )["username"],
        f"{int((datetime.datetime.utcnow() - start).total_seconds() * 1000)}ms",
    )
    return resp


def allowCors(origins):
    @middleware
    async def middle(request, handler):
        resp = await handler(request)
        resp.headers["Access-Control-Allow-Credentials"] = "true"
        for origin in origins:
            resp.headers.add("Access-Control-Allow-Origin", origin)
        return resp

    return middle


def authentication(db, secretKey, cookieName="authentication"):
    @middleware
    async def middle(request, handler):
        try:
            value = request.cookies[cookieName]
            parsed = jwt.decode(
                value,
                secretKey,
                algorithms=['HS256'],
            )
            request.currentUser = await db.user.find_one({
                "_id": bson.ObjectId(parsed["id"]),
            })
        except:
            pass

        return await handler(request)

    return middle


async def setCurrentUser(
    method,
    db,
    secretKey,
    user,
    cookiename="authentication",
    expiration=datetime.timedelta(days=1)
):
    now = datetime.datetime.utcnow()
    token = jwt.encode(
        payload={
            "id": str(user["_id"]),
            "username": user["username"],
            "email": user["email"],
            "exp": now + expiration,
            "iat": now,
        },
        key=secretKey,
        algorithm='HS256',
    ).decode("utf-8")
    method.setCookie("authentication", token)
    return token


from ..errors import ClientError


class ParseError(ClientError):
    def __init__(self, message="Error when parsing request", statusCode=422):
        self.statusCode = statusCode or 422
        self.message = message

    def __repr__(self):
        return self.message



import json
from inspect import getdoc
from aiohttp import web


async def handleDocumentation(handler):
    docs = {"description": getdoc(handler)}

    docs["methods"] = {
        methodName: methodClass().describe()
        for methodName, methodClass in handler.methods().items()
    }

    return web.Response(
        text=json.dumps(docs), headers={"Content-Type": "application/json"}
    )



from ..handler import Method, Endpoint
from ..parsing import (
    RequestParser,
    DictParser,
    ResponseSerializer,
)
from ..parsing.fields import (
    NonBlankStringField,
    EmailField,
    StringField,
    JwtField,
    Field,
)

from ..errors import ServerError
import jwt
import datetime


class HealthCheckEndpoint(Endpoint):
    """
        Return the status of the server
    """

    class Get(Method):
        """
            Return the status of the server
        """


        class Serializer(ResponseSerializer):
            def __init__(self):
                self.healthy = Field(
                    description="A boolean of whether the server is healthy",
                    default=True,
                )

        async def process(
            self,
        ):
            try:
                resp = await self.settings.db.command("ping")

                return {"healthy": True}
            except Exception as e:
                print(e)
                raise ServerError(message="Server is unhealthy")



import asyncio

from .user import createUserCollection
from .authentication import createAuthenticationCollection
from .activity import createActivityCollection


async def createCollections(db):
    await asyncio.wait([
        createUserCollection(db),
        createAuthenticationCollection(db),
        createActivityCollection(db)
    ])